#!/usr/bin/env python
#
# PkgTrackInst.Trackers - Tracks installation log
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42, (c) Poul-Henning Kamp):
# Hye-Shik Chang <perky@FreeBSD.org> wrote this file.  As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some
# day, and you think this stuff is worth it, you can buy me a beer in return.
#
# Hye-Shik Chang
# ----------------------------------------------------------------------------
# $Id: Trackers.py 644 2006-06-25 09:44:16Z perky $

__pychecker__ = 'no-argsused no-isliteral'

import os
from PkgTrackInst import options
from PkgTrackInst.cmd_support import errx


class InstEntry(object):

    root = None             # singleton object for root directory
    abspath = None
    created = False
    realpath = None         # for InstSymbolicLink only

    def __init__(self, created=False):
        self.created = created

    def create_root():
        InstEntry.root = InstDirectory()
        InstEntry.root.abspath = '/'
        return InstEntry.root
    create_root = staticmethod(create_root)

    def traverse(self, depth=0):
        print '  ' * depth + repr(self)

    def iterchildren(self):
        return iter([])

    def walk(self, callback):
        callback(self)
        for name, child in self.iterchildren():
            child.walk(callback)


class InstDirectory(InstEntry):

    def __init__(self, created=False):
        InstEntry.__init__(self, created)
        self.entries = {}

    def __getitem__(self, name):
        pathparts = name.split(os.sep)
        if self.abspath == '/' and not pathparts[0]:
            del pathparts[0]
        if len(pathparts) > 1:
            return self[pathparts[0]][os.sep.join(pathparts[1:])]

        name = pathparts[0]
        if name not in self.entries:
            path = os.path.join(self.abspath, name)
            if os.path.isdir(path):
                self.entries[name] = InstDirectory()
                self.entries[name].abspath = path
            else: # untracked path
                raise KeyError, os.path.join(self.abspath, name)
        return self.entries[name]

    def __setitem__(self, path, obj):
        assert isinstance(obj, InstEntry)

        pathparts = self.split_path(path)
        if len(pathparts) == 1:
            self.entries[pathparts[0]] = obj
            obj.abspath = os.path.join(self.abspath, pathparts[0])
        else:
            self[pathparts[0]][os.sep.join(pathparts[1:])] = obj

    def __delitem__(self, path):
        pathparts = self.split_path(path)
        if len(pathparts) == 1:
            if pathparts[0] not in self.entries:
                return # untracked object
            del self.entries[pathparts[0]]
        else:
            del self[pathparts[0]][os.sep.join(pathparts[1:])]

    def __repr__(self):
        return ('<InstDirectory abspath=%s created=%s children=%s>' % (
                    repr(self.abspath), repr(self.created), 
                    repr(self.entries.keys())))

    def move(self, srcfile, dst):
        assert len(srcfile.split(os.sep)) == 1
        if srcfile not in self.entries:
            # treat it like a new file.
            self.root[dst] = InstFile(created=True)
            return

        sobj = self.entries[srcfile]
        del self.entries[srcfile]
        self.root[dst] = sobj

        def fixup_abspath(newpath, obj):
            obj.abspath = newpath
            for name, child in obj.iterchildren():
                fixup_abspath(os.path.join(newpath, name), child)
        fixup_abspath(dst, sobj)

    def traverse(self, depth=0):
        print '  ' * depth + repr(self)
        for name, obj in self.entries.iteritems():
            print '  ' * depth, name, ':'
            obj.traverse(depth+1)

    def split_path(self, path):
        parts = path.split(os.sep)
        if self.abspath == '/' and not parts[0]:
            del parts[0]
        return parts

    def iterchildren(self):
        return self.entries.iteritems()


class InstFile(InstEntry):

    def __repr__(self):
        return '<InstFile abspath=%s created=%s>' % (repr(self.abspath),
                                                     repr(self.created))


class InstSymbolicLink(InstEntry):

    def __init__(self, realpath, created=False):
        InstEntry.__init__(self, created)
        self.realpath = realpath

    def __repr__(self):
        return '<InstSymbolicLink abspath=%s created=%s realpath=%s>' % (
                repr(self.abspath), repr(self.created), repr(self.realpath))


class MTreeMarker(object):

    def __init__(self, prefix, mtreefile):
        self.prefix = prefix
        self.filepath = mtreefile
        self.selected = []
        for namepart in prefix.split(os.sep):
            if not self.selected:
                self.selected.append(TrackRoot)
            else:
                childnode = self.selected[-1][namepart]
                self.selected.append(childnode)

    def mark(self):
        for line in open(self.filepath):
            sline = line.strip()
            if not sline or sline[0] in '#/':
                continue # we don't need to parse modes/owners/comments

            if sline == '.':
                self.found_new_node(self.selected[-1])
            elif sline == '..':
                self.selected.pop()
            else:
                namepart = sline.split()[0]
                try:
                    childnode = self.selected[-1][namepart]
                except KeyError:
                    childnode = self.selected[-1][namepart] = InstDirectory()
                self.selected.append(childnode)
                self.found_new_node(childnode)

        # Remove after-mtree symlink created by Mk/bsd.port.mk
        prefixnode = TrackRoot[self.prefix]
        try:
            prefixnode['share/nls/POSIX'].created = False
        except KeyError:
            pass

        try:
            prefixnode['share/nls/en_US.US-ASCII'].created = False
        except KeyError:
            pass

    def found_new_node(self, node):
        self.selected[-1].created = False


# singleton for the root directory
TrackRoot = InstEntry.create_root()
